import dataclasses
import re
from urllib.parse import urlparse
from security_engine.models.transaction import Tx
from security_engine.models.action import BaseAction, SignType
from security_engine.models.chain import Chain


@dataclasses.dataclass()
class BaseContext(object):
    action: BaseAction
    origin: str = dataclasses.field(init=False)
    domain: str = dataclasses.field(init=False)
    sign_type: SignType = dataclasses.field(init=False)
    chain: Chain = dataclasses.field(init=False)
    
    def __post_init__(self):
        self.origin = self.get_origin(self.action.origin)
        self.domain = self.get_domain(self.origin)
    
    def get_origin(self, origin):
        if not origin:
            return ''
        return origin.rstrip('/')

    def get_domain(self, origin):
        if not origin:
            return ''
        return urlparse(origin).netloc


@dataclasses.dataclass()
class TextContext(BaseContext):
    text: str = dataclasses.field(init=False)

    def __post_init__(self):
        super(TextContext, self).__post_init__()
        self.text = self.action.text
        self.sign_type = SignType.text

    def match_sign_text(self, text, text_sign_pattern):
        p = re.compile(text_sign_pattern)
        m = p.match(text)
        return True if m else False


@dataclasses.dataclass()
class TransactionContext(BaseContext):
    tx: Tx = dataclasses.field(init=False)
    
    def __post_init__(self):
        super(TransactionContext, self).__post_init__()
        self.tx = self.get_tx(self.action)
        self.sign_type = SignType.transaction

    def get_tx(self, action):
        tx = action.transaction
        if not tx:
            return
        return Tx(**dict(
            chain_id=tx['chainId'],
            data=tx.get('data', '0x'),
            from_=tx.get('from', '').lower(),
            to=tx.get('to', '').lower(),
            gas=int(tx.get('gas', '0x0'), 16),
            gas_price=int(tx.get('gasPrice', '0x0'), 16),
            max_fee_per_gas=int(tx.get('maxFeePerGas', '0x0'), 16),
            max_priority_fee_per_gas=int(tx.get('maxPriorityFeePerGas', '0x0'), 16),
            nonce=int(tx.get('nonce', '0x0'), 16),
            value=int(tx.get('value', '0x0'), 16)
        ))


class DataSource(dict):

    def __getattr__(self, k):
        return self[k]

    @classmethod
    def from_dict(cls, dct: dict):
        ins = cls(**dct)
        return ins


class AddressSet(DataSource):
    pass

AddressSet.__name__ = 'address_set'

class DomainSet(DataSource):
    pass

DomainSet.__name__ = 'domain_set'


class SignTextPatternSet(DataSource):
    pass

SignTextPatternSet.__name__ = 'sign_text_pattern_set'


from enum import Enum
import typing
import dataclasses


class SignType(Enum):
    transaction = 1
    text = 2


@dataclasses.dataclass()
class BaseAction(object):
    sign_type: str = dataclasses.field(init=False)
    origin: str
    
    def __post_init__(self):
        pass

@dataclasses.dataclass()
class TransactionAction(BaseAction):
    chain_id: int = dataclasses.field(init=False)
    transaction: dict

    def __post_init__(self):
        super(TransactionAction, self).__post_init__()
        self.sign_type = SignType.transaction
        self.chain_id = self.transaction['chainId']


@dataclasses.dataclass()
class TextAction(BaseAction):

    text: str
    chain_id: int
        
    def __post_init__(self):
        super(TextAction, self).__post_init__()
        self.sign_type = SignType.text


def get_action(params):
    if SignType.transaction.name in params:
        action = TransactionAction(**params)
    elif SignType.text.name in params:
        action = TextAction(**params)
    else:
        return None
    return action






from security_engine.models.rule import COMMON_ORIGIN
from security_engine.models.data_source import DataSource


class RuleManager(object):

    def __init__(self, load_handlers=[]) -> None:
        self.load_handlers = load_handlers
        self.app_list = []

    def load(self, refresh=True):
        app_list = []
        for load_handler in self.load_handlers:
            app_list.extend(load_handler.load(refresh=refresh))
        self.app_list = app_list
        
    def add_load_handler(self, handler):
        self.load_handlers.append(handler)
    
    def get_data_source_dict(self, app):
        data_source_dict = {}
        for set_name, source_dict in app.data_source.items():
            source = {}
            for k, v in source_dict.items():
                if callable(v):
                    source[k] = v()
                else:
                    source[k] = v
            data_source_dict[set_name] = DataSource.from_dict(source)
        return data_source_dict

    def filter(self, origin=''):
        rule_app_list = []
        for app in self.app_list:
            if app.origin == COMMON_ORIGIN:
                rule_app_list.append(app)
            elif app.origin == origin:
                rule_app_list.append(app)
        return rule_app_list
                





from functools import wraps
from types import FunctionType
from typing import Callable, Any, Tuple

from security_engine.runtime.context import BaseContext


class LogManager(object):
    
    def __init__(self, handlers=[]) -> None:
        self.handlers = handlers

    def load(self):
        pass

    def add_handler(self, handler):
        self.handlers.append(handler)

    def output(self, context):
        for handler in self.handlers:
            handler.output(context)

    def execute(self, fn: FunctionType, *args: Tuple[Any], **kwargs: Any) -> Any:
        res = fn(*args, **kwargs)
        if args and isinstance(args[1], BaseContext):
            self.output(args[1])
        return res

    def __call__(self, fn: FunctionType) -> Callable[..., Any]:

        @wraps(fn)
        def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
            return self.execute(fn, *args, **kwargs)

        return wrapper



import logging
import pprint
import time
from security_engine.handlers import HandlerType, BaseHandler
from security_engine.models.rule import App, ExecuteLog
from security_engine.runtime.context import BaseContext


class BaseLogHandler(BaseHandler):

    @property
    def handler_type(self):
        return HandlerType.LOG

    def output(self):
        raise NotImplemented


class StreamLogHandler(BaseLogHandler):
    
    def __init__(self) -> None:
        super(StreamLogHandler, self).__init__()
        
    def output(self, context: BaseContext):
        app = App(name=context.app.name, 
                is_active=context.app.is_active, 
                version=context.app.version, 
                origin=context.app.origin,
                data_source=None,
                rules=None
                )

        log = ExecuteLog(
                   app=app,
                   hit_rules=context.hit_rules,
                   action=context.action, 
                   time_at=time.time()
                   )
        if hasattr(context, 'error'):
            log.error = context.error
        pprint.pprint(log)
        print('\n')


import dataclasses
from .contract import Contract


@dataclasses.dataclass()
class NFT(object):
    id: str
    name: str
    inner_id: str
    is_erc721: bool
    is_erc1155: bool
    contract: Contract


class NFTCollection(object):
    id: str
    is_erc1155: bool
    is_erc721: bool
    contract: Contract



import dataclasses
import typing


@dataclasses.dataclass()
class Tx(object):
    
    chain_id: int
    from_: str
    to: str
    value: int
    data: bytes
    gas: int
    nonce: int
    gas_price: typing.Optional[int]
    max_fee_per_gas: typing.Optional[int] = 0
    max_priority_fee_per_gas: typing.Optional[int] = 0
    params: typing.Optional[typing.List[dict]] = dataclasses.field(default_factory=list)
    func: typing.Optional[str] = ''




from .address import Address
from .token import Token
from .chain import Chain
from .contract import Contract
from .nft import NFT, NFTCollection
from .rule import Rule, Response
from .transaction import Tx
from .sign_scene import *


import dataclasses
from .chain import Chain


@dataclasses.dataclass()
class Contract(object):
    id: str
    chain: Chain
    code_id: str
    creator: str
    is_token: bool
    is_suicide: bool
    is_multisig: bool




import logging
from security_engine.models.rule import Hit, Response, Level, App
from security_engine.managers.rule import RuleManager
from security_engine.managers.log import LogManager
from security_engine.managers.context import ContextManager
from security_engine.handlers import HandlerType


log_manager = LogManager()


class SecurityEngineCore(object):

    def __init__(self, rule_load_handler_list=[]):
        self.rule_manager = RuleManager(load_handlers=rule_load_handler_list)
        
    def load(self, refresh=True):
        self.rule_manager.load(refresh=refresh)
    
    def add_handler(self, handler):
        if handler.handler_type == HandlerType.RULE_LOAD:
            self.rule_manager.add_load_handler(handler)
        elif handler.handler_type == HandlerType.LOG:
            log_manager.add_handler(handler)
        else:
            logging.error('Engine handler type: %s invalid', handler.handler_type)

    def run(self, context):
        self.context_manager = ContextManager(context=context)
        app_list = self.rule_manager.filter(context.origin)
        hits = []
        for app in app_list:
            data_source_dict = self.rule_manager.get_data_source_dict(app)
            context = self.context_manager.clone(**data_source_dict)
            hit_rules, level = self.run_app(context, app)
            if hit_rules:
                simple_app = App(name=app.name, 
                    is_active=app.is_active, 
                    version=app.version, 
                    origin=app.origin,
                    data_source=None,
                    rules=None
                )
                hit = Hit(app=simple_app, rules=hit_rules, level=level)
                hits.append(hit)
        res = Response(hits=hits)
        return res

    @log_manager
    def run_app(self, context, app):
        hit_rules = []
        level = Level.Safe.value
        for rule in app.rules:
            if rule.sign_type != context.sign_type.value:
                continue
            try:
                hit = self.execute(context, rule)
                if not hit:
                    continue
                if rule.level > level:
                    level = rule.level
                hit_rules.append(rule)
            except Exception as e:
                self.context_manager.add_property(context, 'error', e)
                logging.exception('Eval rule=[%s] error=[%s]', rule.logic, e)
        self.context_manager.add_property(context, 'hit_rules', hit_rules)
        self.context_manager.add_property(context, 'app', app)
        return hit_rules, level

    def execute(self, context, rule):
        context_dict = self.context_manager.to_dict(context)
        if rule.properties:
            properties = {k: eval(v, context_dict) for k, v in rule.properties.items()}
            context_dict.update(properties)
        if rule.conditions:
            for condition in rule.conditions:
                rule.logic = rule.logic.replace(condition['condition'], '(%s)' % condition['logic'])
        return eval(rule.logic, context_dict)


    



import copy

class ContextManager(object):
    
    def __init__(self, context) -> None:
        self.context = context

    @staticmethod
    def add_property(context, key, value):
        setattr(context, key, value)
    
    def clone(self, **kwargs):
        context = copy.deepcopy(self.context)
        for k, v in kwargs.items():
            self.add_property(context, k, v)
        return context
    
    def to_dict(self, context):
        attr_dict = dict()
        for attr in dir(context):
            if not attr.startswith('__'):
                attr_dict[attr] = getattr(context, attr)
        return attr_dict


import dataclasses
from typing import List
from .contract import Contract
from .chain import Chain


@dataclasses.dataclass()
class Address(object):
    id: str
    contract_list: List[Contract] = dataclasses.field(default_factory=list)
    used_chain_list: List[Chain] = dataclasses.field(default_factory=list)







from setuptools import setup, find_packages


setup(
    name="web3-security-engine",
    version="1.2.4",
    author="Debank",
    author_email="security@debank.com",
    description="Web3 Security Engine",
    url="https://debank.com", 
    packages=find_packages(),
    python_requires='>=3.7',
    install_requires=[
        'PyYAML==6.0',
        'smmap==5.0.0',
        'requests==2.27.1'
    ]
)


import dataclasses
from .contract import Contract


@dataclasses.dataclass()
class Token(object):

    id: str
    name: str
    symbol: str
    decimals: int
    total_supply: int
    contract: Contract



import os
import logging
import yaml
import io
import zipfile
import re
import json
import requests
import itertools
import shutil
from security_engine.models.rule import Rule, App, SignType
from security_engine.models.data_source import AddressSet, DomainSet, SignTextPatternSet
from security_engine.handlers import HandlerType, BaseHandler


REPO_TMP_DIR = 'tmp'
RULE_SET_DIR = 'ruleset'


logger = logging.getLogger('rule_load_handler')


class BaseRuleLoadHandler(BaseHandler):
    
    @property
    def handler_type(self):
        return HandlerType.RULE_LOAD

    def fetch_raw(self):
        raise NotImplemented
    
    def load(self):
        raise NotImplemented

        
class FileRuleLoadHandler(BaseRuleLoadHandler):

    def __init__(self, app_list) -> None:
        self.app_list = app_list
        self.sign_type_list = [SignType.text, SignType.transaction]
        self.data_source_class_list = [AddressSet, DomainSet, SignTextPatternSet]

    def fetch_raw(self):
        if not os.path.exists(REPO_TMP_DIR):
            os.mkdir(REPO_TMP_DIR)
        return self.app_list

    def _walk_dir(self, dir_path, callback):
        item = {}
        if os.path.exists(dir_path):
            filenames = [filename for filename in os.listdir(dir_path)]
            for filename in filenames:
                sub_dir_path = os.path.join(dir_path, filename)
                if os.path.isdir(sub_dir_path):
                    objs = self._walk_dir(sub_dir_path, callback)
                else:
                    objs = callback(os.path.join(dir_path, filename))
                item[filename.split('.')[0]] = objs
        return item
    
    def parse_rule(self, app_name):
        rule_list = []
        for sign_type in self.sign_type_list:
            rule_dict = self._walk_dir(os.path.join(REPO_TMP_DIR, app_name, RULE_SET_DIR, sign_type.name), self._load_rules)            
            for rule in itertools.chain(*rule_dict.values()):
                rule.sign_type = sign_type.value
                rule_list.append(rule)
        return rule_list

    def _load_rules(self, file):
        with open(file) as f:
            raw_rule_list = yaml.safe_load(f)
            return [Rule(**raw_rule) for raw_rule in raw_rule_list]
            
    def parse_data_source(self, app_name):
        data_source_dict = {}
        for data_source_class in self.data_source_class_list:
            raw_data_source_dict = self._walk_dir(os.path.join(REPO_TMP_DIR, app_name, data_source_class.__name__), self._load_data_source)
            data_source_dict.update({data_source_class.__name__: raw_data_source_dict})
        return data_source_dict
        
    def _load_data_source(self, file):
        suffix = file.split('.')[-1]
        item = None
        with open(file) as f:
            if suffix == 'json':
                item = json.load(f)
            elif suffix == 'txt':
                if SignTextPatternSet.__name__ in file:
                    item = f.read()
                else:
                    item = []
                    for line in f:
                        if line.startswith('#'):
                            continue
                        l = line.split('#')[0]
                        item.append(l.strip()) if l else None
            elif suffix == 'yaml':
                item = yaml.safe_load(f)
        return item

    def parse(self):
        app_list = []
        for app in self.app_list:
            rule_list = self.parse_rule(app['name'])
            data_source = self.parse_data_source(app['name'])
            app = App(name=app['name'], origin=app['origin'], rules=rule_list, data_source=data_source, is_active=True, version=app['version'])
            app_list.append(app)
        return app_list

    def load(self):
        return self.parse()


class GithubRepoRuleLoadHandler(FileRuleLoadHandler):

    def __init__(self, repo_list, auth_token) -> None:
        self.headers = {'Accept': 'application/vnd.github+json', 'Authorization': auth_token }
        self.repo_list = repo_list
        app_list = self.get_app_list(repo_list)
        super(GithubRepoRuleLoadHandler, self).__init__(app_list)

    @staticmethod
    def get_app_name(repo_url):
        pattern = '(https:\/\/github.com\/)(?P<filepath>.*)'
        match = re.findall(pattern, repo_url)
        if not match:
            raise Exception('Invalid github repo url:%s' % repo_url)
        return match[0][1].replace('/', '_')

    @classmethod
    def get_app_list(cls, repo_list):
        app_list = []
        for repo in repo_list:
            app_list.append(dict(
                name=cls.get_app_name(repo['url']),
                version=repo['commit_hash'],
                origin=repo['origin']
            ))
        return app_list

    def load(self, refresh=True):
        if refresh:
            self.fetch_raw()
        return self.parse()

    def fetch_raw(self):
        super(GithubRepoRuleLoadHandler, self).fetch_raw()
        assert len(self.repo_list) == len(self.app_list)
        for i, repo in enumerate(self.repo_list):
            local_dst = os.path.join(REPO_TMP_DIR, self.app_list[i]['name'])
            self.download_repo_file(repo, local_dst)

    def download_repo_file(self, repo, local_dst):
        if os.path.exists(local_dst):
            shutil.rmtree(local_dst)
        *_, owner, repo_name = repo['url'].split('/')
        
        gen_download_url = "https://api.github.com/repos/{owner}/{repo_name}/zipball/{commit_hash}".format(owner=owner, repo_name=repo_name, commit_hash=repo['commit_hash'])
        archive_url = requests.get(gen_download_url, headers=self.headers).url
        r = requests.get(archive_url)

        zip_ref = zipfile.ZipFile(io.BytesIO(r.content))
        zip_ref.extractall(REPO_TMP_DIR)
        os.rename(os.path.join(REPO_TMP_DIR, zip_ref.namelist()[0]), local_dst)


import dataclasses
import typing
from .token import Token
from .nft import NFT, NFTCollection
from .contract import Contract


@dataclasses.dataclass()
class TokenTransfer(object):
    from_: str
    to: str
    token: Token
    amount: int


@dataclasses.dataclass()
class TokenApprove(object):
    owner: str
    spender: str
    spender_contract: typing.Optional[Contract]
    token: Token
    amount: int


@dataclasses.dataclass()
class NFTTransfer(object):
    from_: str
    to: str
    nft: NFT
    amount: int


@dataclasses.dataclass()
class NFTApprove(object):
    owner: str
    spender: str
    spender_contract: typing.Optional[Contract]
    nft: NFT
    amount: int


@dataclasses.dataclass()
class NFTCollectionApprove(object):
    owner: str
    spender: str
    spender_contract: typing.Optional[Contract]
    nft_collection: NFTCollection
    amount: int


import dataclasses
from enum import Enum
from typing import List, Dict
from .action import BaseAction, SignType

COMMON_ORIGIN = 'common'
DATA_SOURCE = 'data_source'


class Level(Enum):
    Safe = 0
    Warning = 1
    Danger = 2
    Forbidden = 3


@dataclasses.dataclass()
class Condition(object):
    condition: str
    description: str
    logic: str


@dataclasses.dataclass()
class Rule(object):
    description: str
    level: Level
    logic: str
    sign_type: SignType = dataclasses.field(init=False)
    conditions: List[Condition] = dataclasses.field(default_factory=list)
    properties: Dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass()
class App(object):
    name: str
    rules: List[Rule]
    data_source: Dict
    is_active: bool
    version: str
    origin: str = COMMON_ORIGIN


@dataclasses.dataclass()
class Hit(object):
    app: App
    level: str
    rules: List[Rule] = dataclasses.field(default_factory=list)
    

@dataclasses.dataclass()
class Response(object):
    hits: List[Hit] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class ExecuteLog(object):
    app: App
    action: BaseAction
    time_at: int
    hit_rules: List[Rule]
    error: str = None

